#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 31 19:25:08 2021

@author: akshat
"""
import os, copy
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, TensorDataset

import rdkit
from rdkit import Chem
from rdkit import RDLogger
from rdkit.Chem import Descriptors
RDLogger.DisableLog('rdApp.*')

import inspect
from collections import OrderedDict

import multiprocessing
from typing import List
from .features import get_mol_info

def get_mol_feature(smi: str) -> np.array:
    """
    Given a SMILES string (smi), a user needs to provide code for creating descriptors
    for the molecules. This will be used as features for the (1) classifier NN and (2)
    the discriminator (if active).

    This function will be parallelised by processes.

    Parameters
    ----------
    smi : str
        Valid SMILE string generated by JANUS.

    Returns
    -------
    descriptor : array
        Property value of SMILES string 'smi'.

    """
    # mol = Chem.MolFromSmiles(smi)
    # descriptor = AllChem.GetMorganFingerprintAsBitVect(mol, 3, 1024)
    descriptor = get_mol_info(smi)

    return np.array(descriptor)


def obtain_features(smi_list: List[str], num_workers: int = 1):
    assert num_workers <= multiprocessing.cpu_count(), 'num_workers exceed cpu count'
    with multiprocessing.Pool(num_workers) as pool:
        dataset_x = pool.map(
            get_mol_feature, smi_list
        )
        
    return np.array(dataset_x)


class MLP(nn.Module):
    def __init__(self, h_sizes: List[int], n_input: int, n_output: int):
        super(MLP, self).__init__()
        # Layers
        self.hidden = nn.ModuleList([nn.Linear(n_input, h_sizes[0])])
        for k in range(len(h_sizes)-1):
            self.hidden.append(nn.Linear(h_sizes[k], h_sizes[k+1]))
        self.predict = nn.Linear(h_sizes[-1], n_output)

    def forward(self, x):
        for layer in self.hidden:
            x = torch.sigmoid(layer(x))
        output= torch.sigmoid(self.predict(x))
        return output

class EarlyStopping():
    ''' Class that checks criteria for early stopping. Saves the best model weights.
    '''
    def __init__(self, patience, min_delta, mode='minimize'):
        self.patience = patience
        self.best_weights = None
        self.checkpoint = 0
        self.best_epoch = 0
        if mode == 'maximize':
            self.monitor_fn = lambda a, b: np.greater(a - min_delta, b)
            self.best_val = -np.inf
        elif mode == 'minimize':
            self.monitor_fn = lambda a, b: np.less(a + min_delta, b)
            self.best_val = np.inf
        else:
            raise ValueError(f'Mode should be either minimize or maximize.')

    def check_criteria(self, net, epoch, new_val):
        ''' Compare with value in memory. If there is an improvement, reset the checkpoint and
        save the model weights.
        Return True if stopping criteria is met (checkpoint is exceeds patience), otherwise, return False.
        '''
        if self.monitor_fn(new_val, self.best_val):
            self.best_val = new_val
            self.checkpoint = 0
            self.best_weights = copy.deepcopy(net.state_dict())
            self.best_epoch = epoch
        else:
            self.checkpoint += 1
        
        return self.checkpoint > self.patience

    def restore_best(self, net, verbose=True):
        if verbose: print(f'        Early stopping at epoch: {self.best_epoch}       loss: {self.best_val}')
        net.load_state_dict(self.best_weights)
        return net


def get_device(use_gpu: bool):
    if use_gpu:
        device = 'cuda' if torch.cuda.is_available() else 'cpu'   
        if device == 'cpu':
            print('No GPU available, defaulting to CPU.')
    else:
        device = 'cpu'

    return device

############################################################################################

def create_network(n_hidden: List[int], n_input: int, n_output: int, device: str):
    ''' Obtain network
    
    Parameters:
    n_hidden             (list) : Intermediate discrm layers (e.g. [100, 10])
    device               (str)  : Device discrm. will be initialized 
    
    Returns:
    net : torch model
    optimizer   : Loss function optimized (Adam)
    loss_func   : Loss (Cross-Entropy )
    '''
    net = MLP(n_hidden, n_input, n_output).to(device)
    optimizer = torch.optim.Adam(net.parameters(), lr=0.001, weight_decay=1e-4)
    loss_func = nn.BCELoss()
    
    return net, optimizer, loss_func


def train_valid_split(data_x: np.array, data_y: np.array, train_ratio: float = 0.8, 
        seed: int = 30624700):
    ''' Return a random split of training and validation data. 
    Ratio determines the size of training set. Avoids use of sklearn.
    '''
    n = data_x.shape[0]
    train_n = np.floor(n*train_ratio).astype(int)
    indices = np.random.RandomState(seed=seed).permutation(n)
    
    train_x, train_y = data_x[indices[:train_n]], data_y[indices[:train_n]]
    valid_x, valid_y = data_x[indices[train_n:]], data_y[indices[train_n:]]

    return train_x, train_y, valid_x, valid_y


def do_x_training_steps(data_x: np.array, data_y: np.array, net: nn.Module, 
        optimizer: torch.optim.Optimizer, loss_func: nn.Module, 
        steps: int, batch_size: int, device: str):
    ''' Do steps for training. Set batch_size to -1 for full batch training, 
    and 1 for SGD. 
    '''
    train_x, train_y, valid_x, valid_y = train_valid_split(data_x, data_y)
    train_x = torch.tensor(train_x, device=device, dtype=torch.float)
    train_y = torch.tensor(train_y, device=device, dtype=torch.float)
    valid_x = torch.tensor(valid_x, device=device, dtype=torch.float)
    valid_y = torch.tensor(valid_y, device=device, dtype=torch.float)

    if batch_size == -1:
        batch_size = train_x.shape[0]

    train_loader = DataLoader(TensorDataset(train_x, train_y), batch_size=batch_size, shuffle=True)
    valid_loader = DataLoader(TensorDataset(valid_x, valid_y), batch_size=batch_size)

    early_stop = EarlyStopping(patience=500, min_delta=1e-7, mode='minimize')
    net.train()
    for t in range(steps):
        # training steps
        for x, y in train_loader:
            pred_y = net(x)
            loss = loss_func(pred_y, y)

            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

        # validation steps
        val_loss = 0
        net.eval()
        with torch.no_grad():
            for x, y in valid_loader:
                pred_y = net(x)
                loss = loss_func(pred_y, y)
                val_loss += loss
        net.train()
        val_loss /= len(valid_loader)

        if t % 1000 == 0: 
            print('        Epoch:{} Loss:{}'.format(t, loss.item()))
            print(f'                Validation loss: {val_loss.item()}')

        # check early stopping criteria
        stop = early_stop.check_criteria(net, t, val_loss.item())
        if stop:
            net = early_stop.restore_best(net)
            break

    return net


def create_and_train_network(smi_list: List[str], targets: np.array, n_hidden: List[int] = [100, 10], 
        use_gpu: bool = True, num_workers: int = 1): 
    ''' Featurize smiles and train classifier network. Return trained network.
    '''
    # featurize
    dataset_x = obtain_features(smi_list, num_workers=num_workers)
    avg_val = np.percentile(targets, 80) # np.average(targets)
    dataset_y  = np.expand_dims([1 if x >= avg_val else 0 for x in targets ], -1)
    
    # create network
    device = get_device(use_gpu)
    net, optimizer, loss_func = create_network(n_hidden, dataset_x.shape[-1], dataset_y.shape[-1], device)

    # train network
    net = do_x_training_steps(
        data_x=dataset_x, data_y=dataset_y, net=net, 
        optimizer=optimizer,  loss_func=loss_func, steps=20000, 
        batch_size=1024, device=device
    )
    
    return net


def obtain_model_pred(smi_list: List[str], net: nn.Module, use_gpu: bool = True,
        num_workers: int = 1, batch_size: int = 1024): 
    predictions = []

    device = get_device(use_gpu)
    data_x = obtain_features(smi_list, num_workers=num_workers)
    data_x = torch.tensor(data_x, device=device, dtype=torch.float)
    loader = DataLoader(TensorDataset(data_x), batch_size=batch_size)

    num_batches = -(-data_x.shape[0] // batch_size)     # ceil division

    print('Number of batches: ', num_batches)
    with torch.no_grad():
        for i, x in enumerate(loader): 
            print('        Predicting Batch: {}/{}'.format(i, num_batches))
            outputs = net(x[0])
            out_    = outputs.detach().cpu().numpy()
            predictions.append(out_)
    
    predictions = np.concatenate(predictions, axis=0)   # concatenate in the batch axis

    return predictions




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 31 12:15:57 2021

@author: akshat
"""
from typing import Dict
import random
import multiprocessing

import rdkit
from rdkit import Chem

import selfies 
from selfies import encoder, decoder

from .utils import get_selfies_chars

def mutate_sf(sf_chars, alphabet, num_sample_frags, base_alphabet = None):
    """
    Given a list of SELFIES alphabets, make random changes to the molecule using 
    alphabet. Opertations to molecules are character replacements, additions and deletions. 

    Parameters
    ----------
    sf_chars : (list of string alphabets)
        List of string alphabets for a SELFIE string.
    alphabet : (list of SELFIE strings)
        New SELFIES characters are added here and sampled.
    num_sample_frags: (int)
        Number of randomly sampled SELFIE strings.
    base_alphabet: (list of SELFIE strings)
        Main alphabet that will be appended with the introduced characters above.
        If none, use the semantic robust alphabet.

    Returns
    -------
    Mutated SELFIE string.

    """
    if base_alphabet is None:
        base_alphabet = list(selfies.get_semantic_robust_alphabet())
    random_char_idx = random.choice(range(len(sf_chars)))
    choices_ls = [1, 2, 3]  # 1 = replacement; 2 = addition; 3=delete
    mutn_choice = choices_ls[
        random.choice(range(len(choices_ls)))
    ]  # Which mutation to do:

    if alphabet != []:
        alphabet = random.sample(alphabet, num_sample_frags) + base_alphabet
    else:
        alphabet = base_alphabet

    # Mutate character:
    if mutn_choice == 1:
        random_char = alphabet[random.choice(range(len(alphabet)))]
        change_sf = (
            sf_chars[0:random_char_idx]
            + [random_char]
            + sf_chars[random_char_idx + 1 :]
        )

    # add character:
    elif mutn_choice == 2:
        random_char = alphabet[random.choice(range(len(alphabet)))]
        change_sf = (
            sf_chars[0:random_char_idx] + [random_char] + sf_chars[random_char_idx:]
        )

    # delete character:
    elif mutn_choice == 3:
        if len(sf_chars) != 1:
            change_sf = sf_chars[0:random_char_idx] + sf_chars[random_char_idx + 1 :]
        else:
            change_sf = sf_chars

    return "".join(x for x in change_sf)


def mutate_smiles(
    smile, alphabet, num_random_samples, num_mutations, num_sample_frags, base_alphabet = None
):
    """
    Given an input smile, perform mutations to the strucutre using provided SELFIE
    alphabet list. 'num_random_samples' number of different SMILES orientations are 
    considered & total 'num_mutations' are performed. 

    Parameters
    ----------
    smile : (str)
        Valid SMILES string.
    alphabet : (list of str)
        list of SELFIE strings.
    num_random_samples : (int)
        Number of different SMILES orientations to be formed for the input smile.
    num_mutations : TYPE
        Number of mutations to perform on each of different orientations SMILES.
    num_sample_frags: (int)
        Number of randomly sampled SELFIE strings.

    Returns
    -------
    mutated_smiles_canon : (list of strings)
        List of unique molecules produced from mutations.
    """
    mol = Chem.MolFromSmiles(smile)
    Chem.Kekulize(mol)

    # Obtain randomized orderings of the SMILES:
    randomized_smile_orderings = []
    for _ in range(num_random_samples):
        randomized_smile_orderings.append(
            rdkit.Chem.MolToSmiles(
                mol,
                canonical=False,
                doRandom=True,
                isomericSmiles=False,
                kekuleSmiles=True,
            )
        )

    # Convert all the molecules to SELFIES
    selfies_ls = [encoder(x) for x in randomized_smile_orderings]
    selfies_ls_chars = [get_selfies_chars(selfie) for selfie in selfies_ls]

    # Obtain the mutated selfies
    mutated_sf = []
    for sf_chars in selfies_ls_chars:

        for i in range(num_mutations):
            if i == 0:
                mutated_sf.append(mutate_sf(sf_chars, alphabet, num_sample_frags, base_alphabet))
            else:
                mutated_sf.append(
                    mutate_sf(
                        get_selfies_chars(mutated_sf[-1]), alphabet, num_sample_frags, base_alphabet
                    )
                )

    mutated_smiles = [decoder(x) for x in mutated_sf]
    mutated_smiles_canon = []
    for item in mutated_smiles:
        try:
            smi_canon = Chem.MolToSmiles(
                Chem.MolFromSmiles(item, sanitize=True),
                isomericSmiles=False,
                canonical=True,
            )
            if smi_canon != "": 
                mutated_smiles_canon.append(smi_canon)
        except:
            continue

    mutated_smiles_canon = list(set(mutated_smiles_canon))
    return mutated_smiles_canon



if __name__ == "__main__":
    molecules_here = [
        "CCC",
        "CCCC",
        "CCCCC",
        "CCCCCCCC",
        "CS",
        "CSSS",
        "CSSSSS",
        "CF",
        "CI",
        "CBr",
        "CSSSSSSSSSSSS",
        "CSSSSSSSSSC",
        "CSSSSCCSSSC",
        "CSSSSSSSSSF",
        "SSSSSC",
    ]
    A = get_mutated_smiles(
        molecules_here, alphabet=["[C]"] * 500, num_sample_frags=200, space="Explore"
    )




import numpy as np
import inspect
import rdkit
from rdkit import Chem
from rdkit import RDLogger
from rdkit.Chem import Descriptors
from collections import OrderedDict
RDLogger.DisableLog('rdApp.*')

def get_rot_bonds_posn(mol):
    '''Return atom indices with Rotatable bonds 
    
    Examples:
    >>> get_rot_bonds_posn('CC1=CC=CC=C1')  # Toluene  (Rotatable Bonds At: CH3 & Benzene)
    ((0, 1),)
    >>> get_rot_bonds_posn('CCC1=CC=CC=C1') # (Rotatable Bonds At: CH3, CH3 & Benzene)
    ((0, 1), (1, 2))
    '''
    RotatableBond = Chem.MolFromSmarts('*-&!@*')
    rot = mol.GetSubstructMatches(RotatableBond)
    return rot

def get_bond_indeces(mol, rot):
    '''Get all the bond indices with Rotatable bonds atoms (generated from 'get_rot_bonds_posn')
    '''
    bonds_idx = []
    for i in range(len(rot)):
        bond = mol.GetBondBetweenAtoms(rot[i][0],rot[i][1])
        bonds_idx.append(bond.GetIdx())
    return bonds_idx


def obtain_rings(smi):
    '''Obtain a list of all rings present in SMILE string smi
    
    Examples:
    >>> obtain_rings('CCC1=CC=CC=C1')
    ['c1ccccc1']
    >>> obtain_rings('C1=CC=C(C=C1)C1=CC=CC=C1')
    ['c1ccccc1', 'c1ccccc1']
    >>> obtain_rings('C1=CC2=C(C=C1)C=CC=C2')
    (None, None)
    
    Parameters:
    smi (string) : SMILE string of a molecule 
    
    Returns
    (list)       : List if all rings in a SMILE string 
    '''
    mol = Chem.MolFromSmiles(smi)
    rot = get_rot_bonds_posn(mol) # Get rotatble bond positions
    
    if len(rot) == 0:
        return None, None
    
    bond_idx = get_bond_indeces(mol, rot)
    new_mol = Chem.FragmentOnBonds(mol, bond_idx, addDummies=False) 
    new_smile = Chem.MolToSmiles(new_mol)
    
    smile_split_list = new_smile.split(".") 
    ring_smiles = []
    for item in smile_split_list:
        if '1' in item and Chem.MolFromSmiles(item) is not None:
            ring_smiles.append(item)
    
    return ring_smiles 


def count_atoms(mol, atomic_num):
    '''Count the number of atoms in mol with atomic number atomic_num
    
    
    Parameters:
    mol (rdkit.Chem.rdchem.Mol) : Molecule in which search is conducted
    atomic_num            (int) : Counting is done in mol for atoms with this atomic number
    Returns:
    (int) :  final count of atom
    '''
    pat = Chem.MolFromSmarts("[#{}]".format(atomic_num))
    return len(mol.GetSubstructMatches(pat))

def get_num_bond_types(mol):
    '''Calculate the ratio of total number of  (single, double, triple, aromatic) bonds to the 
       total number of bonds. 
    
    Parameters:
    mol (rdkit.Chem.rdchem.Mol) : Molecule for which ratios arre retuned 
    
    Returns:
    (list):  [num_single/num_bonds, num_double/num_bonds, num_triple/num_bonds, num_aromatic/num_bonds]
    '''
    bonds = mol.GetBonds()    
    
    num_bonds    = 0
    num_double   = 0
    num_triple   = 0
    num_single   = 0
    num_aromatic = 0
    
    for b in bonds:
        num_bonds += 1
        if b.GetBondType() == rdkit.Chem.rdchem.BondType.SINGLE:
            num_single += 1
        if b.GetBondType() == rdkit.Chem.rdchem.BondType.DOUBLE:
            num_double += 1
        if b.GetBondType() == rdkit.Chem.rdchem.BondType.TRIPLE:
            num_triple += 1
        if b.GetBondType() == rdkit.Chem.rdchem.BondType.AROMATIC:
            num_aromatic += 1
    if num_bonds == 0:
        return [0, 0, 0, 0]
    else:
        return [num_single/num_bonds, num_double/num_bonds, num_triple/num_bonds, num_aromatic/num_bonds]


def count_conseq_double(mol):
    '''Return the number of consequtive double bonds in an entire molecule
       including rings 
    Examples 
    >>> count_conseq_double(Chem.MolFromSmiles('C1=CC=C=C=C1'))
    2
    >>> count_conseq_double(Chem.MolFromSmiles('C1=CC=CC=C1'))
    0
    >>> count_conseq_double(Chem.MolFromSmiles('C1=CC2=C(C=C1)C=C=C=C2'))
    2
    
    Parameters:
    mol (rdkit.Chem.rdchem.Mol) : Molecule for conseq. double bonds are to be counted 
    
    Returns:
    (int):  The integer number of coseq. double bonds 
    '''
    bonds = mol.GetBonds()    
    
    previous_BType    = None
    count_conseq_doub = 0
    for b in bonds:
        curr_BType = b.GetBondType()
        if previous_BType == curr_BType and curr_BType == rdkit.Chem.rdchem.BondType.DOUBLE:
            count_conseq_doub += 1
        previous_BType = curr_BType
    
    return count_conseq_doub


def size_ring_counter(ring_ls):
    '''Get the number of rings of sizes 3 to 20 and the number of consequtive double bonds in a ring
    
    Parameters:
    ring_ls (list)  : list of rings of a molecule 
    
    Returns
    (list)          : Of size 19 (1 for number of conseq. double bonds)
                                 (18 for number of rings between size 3 to 20)
    '''
    ring_counter = []
    
    if ring_ls == (None, None): # Presence of no rings, return 0s for the 19 feature
        return [0 for i in range(19)] 
    
    mol_ring_ls  = [Chem.MolFromSmiles(smi) for smi in ring_ls]
    
    # Cont number consequtive double bonds in ring 
    conseq_dbl_bnd_in_ring = 0
    for item in mol_ring_ls:
        conseq_dbl_bnd_in_ring += count_conseq_double(item)
    ring_counter.append(conseq_dbl_bnd_in_ring) # concatenate onto list ring_counter
    
    # Count size of rings
    for i in range(3, 21):
        count = 0
        for mol_ring in mol_ring_ls:
            if mol_ring.GetNumAtoms() == i:
                count += 1
        ring_counter.append(count)
    return ring_counter

    
def get_mol_info(smi):                 
    ''' Calculate a set of 51 RdKit properties, collected from above helper functions. 
    
    Parameters:
    smi (string) : SMILE string of molecule 
    
    Returns:
    (list of float) : list of 51 calculated properties  
    '''
    mol = Chem.MolFromSmiles(smi)
        
    num_atoms   = mol.GetNumAtoms()       
    num_hydro   = Chem.AddHs(mol).GetNumAtoms() - num_atoms 
    num_carbon  = count_atoms(mol, 6)
    num_nitro   = count_atoms(mol, 7)
    num_sulphur = count_atoms(mol, 16)
    num_oxy     = count_atoms(mol, 8)
    num_clorine = count_atoms(mol, 17)
    num_bromine = count_atoms(mol, 35)
    num_florine = count_atoms(mol, 9)

    
    if num_carbon == 0: # Avoid division by zero error, set num_carbon to a very small value 
        num_carbon = 0.0001
    
    basic_props = [num_atoms/num_carbon, num_hydro/num_carbon, num_nitro/num_carbon, 
                     num_sulphur/num_carbon, num_oxy/num_carbon, num_clorine/num_carbon,
                     num_bromine/num_carbon, num_florine/num_carbon]
    
    to_calculate = ["RingCount", "HallKierAlpha", "BalabanJ", "NumAliphaticCarbocycles","NumAliphaticHeterocycles",
                   "NumAliphaticRings","NumAromaticCarbocycles","NumAromaticHeterocycles",
                   "NumAromaticRings","NumHAcceptors","NumHDonors","NumHeteroatoms",
                   "NumRadicalElectrons","NumSaturatedCarbocycles","NumSaturatedHeterocycles",
                   "NumSaturatedRings","NumValenceElectrons"]    

    # Calculate all propoerties listed in 'to_calculate'
    calc_props = OrderedDict(inspect.getmembers(Descriptors, inspect.isfunction))
    for key in list(calc_props.keys()):
        if key.startswith('_'):
            del calc_props[key]
            continue
        if len(to_calculate)!=0 and key not in to_calculate:
            del calc_props[key]
    # features = [val(mol) for key,val in calc_props.items()] # List of properties 

    features = []
    for key, val in calc_props.items():
        try:
            features.append(val(mol))
        except:
            print(f'Failed at: {val}')
            print(f'Failed at: {Chem.MolToSmiles(mol)}')
    
    
    # Ratio of total number of  (single, double, triple, aromatic) bonds to the total number of bonds. 
    simple_bond_info = get_num_bond_types(mol) 
    
    # Obtain all rings in a molecule and calc. #of triple bonds in rings & #of rings in molecule 
    ring_ls = obtain_rings(smi)
    num_triple = 0      # num triple bonds in ring
    
    if len(ring_ls) > 0 and ring_ls != (None, None): 
        for item in ring_ls:
            num_triple += item.count('#')
        simple_bond_info.append(len(ring_ls))     # append number of Rings in molecule 
    else:   
        simple_bond_info.append(0)                # no rotatable bonds

    simple_bond_info.append(num_triple)          # number of triple bonds in rings

                                              
    # Calculate the number of rings of size 3 to 20 & number of conseq. double bonds in rings 
    simple_bond_info = simple_bond_info + size_ring_counter(ring_ls)
    
    # Calculate the number of consequitve double bonds in entire molecule
    simple_bond_info.append(count_conseq_double(mol)) 
    
    return np.array(features + basic_props + simple_bond_info)


import selfies as sf
import yaml
from rdkit.Chem import AllChem
from rdkit.Chem import MolFromSmiles as smi2mol
from rdkit.Chem import MolToSmiles as mol2smi
from rdkit.DataStructs.cDataStructs import TanimotoSimilarity

def get_selfies_chars(selfies):
    """Obtain a list of all selfie characters in string selfies
    
    Parameters: 
    selfie (string) : A selfie string - representing a molecule 
    
    Example: 
    >>> get_selfies_chars('[C][=C][C][=C][C][=C][Ring1][Branch1_1]')
    ['[C]', '[=C]', '[C]', '[=C]', '[C]', '[=C]', '[Ring1]', '[Branch1_1]']
    
    Returns
    -------
    chars_selfies (list of strings) : 
        list of selfie characters present in molecule selfie
    """
    chars_selfies = sf.split_selfies(selfies)
    return list(chars_selfies)

def sanitize_smiles(smi):
    """
    Return a canonical smile representation of smi 

    Parameters
    ----------
    smi : str
        smile string to be canonicalized 

    Returns
    -------
    mol (rdkit.Chem.rdchem.Mol) : 
        RdKit mol object (None if invalid smile string smi)
    smi_canon (string)          : 
        Canonicalized smile representation of smi (None if invalid smile string smi)
    conversion_successful (bool): 
        True/False to indicate if conversion was  successful 
    """
    if smi == '':
        return None
    try:
        mol = smi2mol(smi, sanitize=True)
        smi_canon = mol2smi(mol, isomericSmiles=False, canonical=True)
        return smi_canon
    except:
        return None

def get_fp_scores(smiles_back, target_smi):
    """
    Given a list of SMILES (smiles_back), tanimoto similarities are calculated 
    (using Morgan fingerprints) to SMILES (target_smi). 
    Parameters
    ----------
    smiles_back : (list)
        List of valid SMILE strings. 
    target_smi : (str)
        Valid SMILES string. 
    Returns
    -------
    smiles_back_scores : (list of floats)
        List of fingerprint similarity scores of each smiles in input list. 
    """
    smiles_back_scores = []
    target = smi2mol(target_smi)
    fp_target = AllChem.GetMorganFingerprint(target, 2)
    for item in smiles_back:
        mol = smi2mol(item)
        fp_mol = AllChem.GetMorganFingerprint(mol, 2)
        score = TanimotoSimilarity(fp_mol, fp_target)
        smiles_back_scores.append(score)
    return smiles_back_scores

def from_yaml(work_dir, 
        fitness_function, 
        start_population,
        yaml_file, **kwargs):

    # create dictionary with parameters defined by yaml file 
    with open(yaml_file, 'r') as f:
        params = yaml.load(f, Loader=yaml.SafeLoader)
    params.update(kwargs)
    params.update({
        'work_dir': work_dir,
        'fitness_function': fitness_function,
        'start_population': start_population
    })

    return params



import os, sys
import multiprocessing
import random
import yaml
from functools import partial
from collections import OrderedDict
from typing import Callable, List, Optional

import numpy as np

from .crossover import crossover_smiles
from .mutate import mutate_smiles
from .network import create_and_train_network, obtain_model_pred
from .utils import sanitize_smiles, get_fp_scores
from .fragment import form_fragments


class JANUS:
    """ JANUS class for genetic algorithm applied on SELFIES
    string representation.
    See example/example.py for descriptions of parameters
    """

    def __init__(
        self,
        work_dir: str,
        fitness_function: Callable,
        start_population: str,
        verbose_out: Optional[bool] = False,
        custom_filter: Optional[Callable] = None,
        alphabet: Optional[List[str]] = None,
        use_gpu: Optional[bool] = True,
        num_workers: Optional[int] = None,
        generations: Optional[int] = 200,
        generation_size: Optional[int] = 5000,
        num_exchanges: Optional[int] = 5,
        use_fragments: Optional[bool] = True,
        num_sample_frags: Optional[int] = 200,
        use_classifier: Optional[bool] = True,
        explr_num_random_samples: Optional[int] = 5,
        explr_num_mutations: Optional[int] = 5,
        crossover_num_random_samples: Optional[int] = 1,
        exploit_num_random_samples: Optional[int] = 400,
        exploit_num_mutations: Optional[int] = 400,
        top_mols: Optional[int] = 1
    ):

        # set all class variables
        self.work_dir = work_dir
        self.fitness_function = fitness_function
        self.start_population = start_population
        self.verbose_out = verbose_out
        self.custom_filter = custom_filter
        self.alphabet = alphabet
        self.use_gpu = use_gpu
        self.num_workers = num_workers if num_workers is not None else multiprocessing.cpu_count()
        self.generations = generations
        self.generation_size = generation_size
        self.num_exchanges = num_exchanges
        self.use_fragments = use_fragments
        self.num_sample_frags = num_sample_frags
        self.use_classifier = use_classifier
        self.explr_num_random_samples = explr_num_random_samples
        self.explr_num_mutations = explr_num_mutations
        self.crossover_num_random_samples = crossover_num_random_samples
        self.exploit_num_random_samples = exploit_num_random_samples
        self.exploit_num_mutations = exploit_num_mutations
        self.top_mols = top_mols

        # create dump folder
        if not os.path.isdir(f"./{self.work_dir}"):
            os.mkdir(f"./{self.work_dir}")
        self.save_hyperparameters()

        # get initial population
        init_smiles, init_fitness = [], []
        with open(self.start_population, "r") as f:
            for line in f:
                line = sanitize_smiles(line.strip())
                if line is not None:
                    init_smiles.append(line)
        # init_smiles = list(set(init_smiles)) 

        # check that parameters are valid
        assert (
            len(init_smiles) >= self.generation_size
        ), "Initial population smaller than generation size."
        assert (
            self.top_mols <= self.generation_size
        ), "Number of top molecules larger than generation size."

        # make fragments from initial smiles
        self.frag_alphabet = []
        if self.use_fragments:
            with multiprocessing.Pool(self.num_workers) as pool:
                frags = pool.map(form_fragments, init_smiles)
            frags = self.flatten_list(frags)
            print(f"    Unique and valid fragments generated: {len(frags)}")
            self.frag_alphabet.extend(frags)

        # get initial fitness
        # with multiprocessing.Pool(self.num_workers) as pool:
        #     init_fitness = pool.map(self.fitness_function, init_smiles)
        init_fitness = []
        for smi in init_smiles:
            init_fitness.append(self.fitness_function(smi))

        # sort the initial population and save in class
        idx = np.argsort(init_fitness)[::-1]
        init_smiles = np.array(init_smiles)[idx]
        init_fitness = np.array(init_fitness)[idx]
        self.population = init_smiles[: self.generation_size]
        self.fitness = init_fitness[: self.generation_size]

        with open(os.path.join(self.work_dir, "init_mols.txt"), "w") as f:
            f.writelines([f"{x}\n" for x in self.population])

        # store in collector, deal with duplicates
        self.smiles_collector = {}
        uniq_pop, idx, counts = np.unique(
            self.population, return_index=True, return_counts=True
        )
        for smi, count, i in zip(uniq_pop, counts, idx):
            self.smiles_collector[smi] = [self.fitness[i], count]

    def mutate_smi_list(self, smi_list: List[str], space="local"):
        # parallelized mutation function
        if space == "local":
            num_random_samples = self.exploit_num_random_samples
            num_mutations = self.exploit_num_mutations
        elif space == "explore":
            num_random_samples = self.explr_num_random_samples
            num_mutations = self.explr_num_mutations
        else:
            raise ValueError('Invalid space, choose "local" or "explore".')

        smi_list = smi_list * num_random_samples
        with multiprocessing.Pool(self.num_workers) as pool:
            mut_smi_list = pool.map(
                partial(
                    mutate_smiles,
                    alphabet=self.frag_alphabet,
                    num_random_samples=1,
                    num_mutations=num_mutations,
                    num_sample_frags=self.num_sample_frags,
                    base_alphabet=self.alphabet
                ),
                smi_list,
            )
        mut_smi_list = self.flatten_list(mut_smi_list)
        return mut_smi_list

    def crossover_smi_list(self, smi_list: List[str]):
        # parallelized crossover function
        with multiprocessing.Pool(self.num_workers) as pool:
            cross_smi = pool.map(
                partial(
                    crossover_smiles,
                    crossover_num_random_samples=self.crossover_num_random_samples,
                ),
                smi_list,
            )
        cross_smi = self.flatten_list(cross_smi)
        return cross_smi

    def check_filters(self, smi_list: List[str]):
        if self.custom_filter is not None:
            smi_list = [smi for smi in smi_list if self.custom_filter(smi)]
        return smi_list

    def save_hyperparameters(self):
        hparams = {
            k: v if not callable(v) else v.__name__ for k, v in vars(self).items()
        }
        with open(os.path.join(self.work_dir, "hparams.yml"), "w") as f:
            yaml.dump(hparams, f)

    def run(self):
        """ Run optimization based on hyperparameters initialized
        """

        for gen_ in range(self.generations):

            # bookkeeping
            if self.verbose_out:
                output_dir = os.path.join(self.work_dir, f"{gen_}_DATA")
                if not os.path.isdir(output_dir):
                    os.mkdir(output_dir)

            print(f"On generation {gen_}/{self.generations}")

            keep_smiles, replace_smiles = self.get_good_bad_smiles(
                self.fitness, self.population, self.generation_size
            )
            replace_smiles = list(set(replace_smiles))

            ### EXPLORATION ###
            # Mutate and crossover (with keep_smiles) molecules that are meant to be replaced
            explr_smiles = []
            timeout_counter = 0
            while len(explr_smiles) < self.generation_size-len(keep_smiles):
                # Mutations:
                mut_smi_explr = self.mutate_smi_list(
                    replace_smiles[0 : len(replace_smiles) // 2], space="explore"
                )
                mut_smi_explr = self.check_filters(mut_smi_explr)

                # Crossovers:
                smiles_join = []
                for item in replace_smiles[len(replace_smiles) // 2 :]:
                    smiles_join.append(item + "xxx" + random.choice(keep_smiles))
                cross_smi_explr = self.crossover_smi_list(smiles_join)
                cross_smi_explr = self.check_filters(cross_smi_explr)

                # Combine and get unique smiles not yet found
                all_smiles = list(set(mut_smi_explr + cross_smi_explr))
                for x in all_smiles:
                    if x not in self.smiles_collector:
                        explr_smiles.append(x)
                explr_smiles = list(set(explr_smiles))

                timeout_counter += 1
                if timeout_counter % 100 == 0:
                    print(f'Exploration: {timeout_counter} iterations of filtering. \
                    Filter may be too strict, or you need more mutations/crossovers.')

            # Replace the molecules with ones in exploration mutation/crossover
            if not self.use_classifier or gen_ == 0:
                replaced_pop = random.sample(
                    explr_smiles, self.generation_size - len(keep_smiles)
                )
            else:
                # The sampling needs to be done by the neural network!
                print("    Training classifier neural net...")
                train_smiles, targets = [], []
                for item in self.smiles_collector:
                    train_smiles.append(item)
                    targets.append(self.smiles_collector[item][0])
                net = create_and_train_network(
                    train_smiles,
                    targets,
                    num_workers=self.num_workers,
                    use_gpu=self.use_gpu,
                )

                # Obtain predictions on unseen molecules:
                print("    Obtaining Predictions")
                new_predictions = obtain_model_pred(
                    explr_smiles,
                    net,
                    num_workers=self.num_workers,
                    use_gpu=self.use_gpu,
                )
                sorted_idx = np.argsort(np.squeeze(new_predictions))[::-1]
                replaced_pop = np.array(explr_smiles)[
                    sorted_idx[: self.generation_size - len(keep_smiles)]
                ].tolist()

            # Calculate actual fitness for the exploration population
            self.population = keep_smiles + replaced_pop
            self.fitness = []
            for smi in self.population:
                if smi in self.smiles_collector:
                    # if already calculated, use the value from smiles collector
                    self.fitness.append(self.smiles_collector[smi][0])
                    self.smiles_collector[smi][1] += 1
                else:
                    # make a calculation
                    f = self.fitness_function(smi)
                    self.fitness.append(f)
                    self.smiles_collector[smi] = [f, 1]

            # Print exploration data
            idx_sort = np.argsort(self.fitness)[::-1]
            print(f"    (Explr) Top Fitness: {self.fitness[idx_sort[0]]}")
            print(f"    (Explr) Top Smile: {self.population[idx_sort[0]]}")

            fitness_sort = np.array(self.fitness)[idx_sort]
            if self.verbose_out:
                with open(
                    os.path.join(
                        self.work_dir, str(gen_) + "_DATA", "fitness_explore.txt"
                    ),
                    "w",
                ) as f:
                    f.writelines(["{} ".format(x) for x in fitness_sort])
                    f.writelines(["\n"])
            else:
                with open(os.path.join(self.work_dir, "fitness_explore.txt"), "w") as f:
                    f.writelines(["{} ".format(x) for x in fitness_sort])
                    f.writelines(["\n"])

            # this population is sort by modified fitness, if active
            population_sort = np.array(self.population)[idx_sort]
            if self.verbose_out:
                with open(
                    os.path.join(
                        self.work_dir, str(gen_) + "_DATA", "population_explore.txt"
                    ),
                    "w",
                ) as f:
                    f.writelines(["{} ".format(x) for x in population_sort])
                    f.writelines(["\n"])
            else:
                with open(
                    os.path.join(self.work_dir, "population_explore.txt"), "w"
                ) as f:
                    f.writelines(["{} ".format(x) for x in population_sort])
                    f.writelines(["\n"])

            ### EXPLOITATION ###
            # Conduct local search on top-n molecules from population, mutate and do similarity search
            exploit_smiles = []
            timeout_counter = 0
            while len(exploit_smiles) < self.generation_size:
                smiles_local_search = population_sort[0 : self.top_mols]
                mut_smi_loc = self.mutate_smi_list(smiles_local_search, "local")
                mut_smi_loc = self.check_filters(mut_smi_loc)

                # filter out molecules already found
                for x in mut_smi_loc:
                    if x not in self.smiles_collector:
                        exploit_smiles.append(x)

                timeout_counter += 1
                if timeout_counter % 100 == 0:
                    print(f'Exploitation: {timeout_counter} iterations of filtering. \
                    Filter may be too strict, or you need more mutations/crossovers.')

            # sort by similarity, only keep ones similar to best
            fp_scores = get_fp_scores(exploit_smiles, population_sort[0])
            fp_sort_idx = np.argsort(fp_scores)[::-1][: self.generation_size]
            # highest fp_score idxs
            self.population_loc = np.array(exploit_smiles)[
                fp_sort_idx
            ]  # list of smiles with highest fp scores

            # STEP 4: CALCULATE THE FITNESS FOR THE LOCAL SEARCH:
            # Exploitation data generated from similarity search is measured with fitness function
            self.fitness_loc = []
            for smi in self.population_loc:
                f = self.fitness_function(smi)
                self.fitness_loc.append(f)
                self.smiles_collector[smi] = [f, 1]

            # List of original local fitness scores
            idx_sort = np.argsort(self.fitness_loc)[
                ::-1
            ]  # index of highest to lowest fitness scores
            print(f"    (Local) Top Fitness: {self.fitness_loc[idx_sort[0]]}")
            print(f"    (Local) Top Smile: {self.population_loc[idx_sort[0]]}")

            fitness_sort = np.array(self.fitness_loc)[idx_sort]
            if self.verbose_out:
                with open(
                    os.path.join(
                        self.work_dir, str(gen_) + "_DATA", "fitness_local_search.txt"
                    ),
                    "w",
                ) as f:
                    f.writelines(["{} ".format(x) for x in fitness_sort])
                    f.writelines(["\n"])
            else:
                with open(
                    os.path.join(self.work_dir, "fitness_local_search.txt"), "w"
                ) as f:
                    f.writelines(["{} ".format(x) for x in fitness_sort])
                    f.writelines(["\n"])

            population_sort = np.array(self.population_loc)[idx_sort]
            if self.verbose_out:
                with open(
                    os.path.join(
                        self.work_dir,
                        str(gen_) + "_DATA",
                        "population_local_search.txt",
                    ),
                    "w",
                ) as f:
                    f.writelines(["{} ".format(x) for x in population_sort])
                    f.writelines(["\n"])
            else:
                with open(
                    os.path.join(self.work_dir, "population_local_search.txt"), "w"
                ) as f:
                    f.writelines(["{} ".format(x) for x in population_sort])
                    f.writelines(["\n"])

            # STEP 5: EXCHANGE THE POPULATIONS:
            # Introduce changes to 'fitness' & 'population'
            best_smi_local = population_sort[0 : self.num_exchanges]
            best_fitness_local = fitness_sort[0 : self.num_exchanges]

            # But will print the best fitness values in file
            idx_sort = np.argsort(self.fitness)[
                ::-1
            ]  # sorted indices for the entire population
            worst_indices = idx_sort[
                -self.num_exchanges :
            ]  # replace worst ones with the best ones
            for i, idx in enumerate(worst_indices):
                try:
                    self.population[idx] = best_smi_local[i]
                    self.fitness[idx] = best_fitness_local[i]
                except:
                    continue

            # Save best of generation!:
            fit_all_best = np.argmax(self.fitness)

            # write best molecule with best fitness
            with open(
                os.path.join(self.work_dir, "generation_all_best.txt"), "a+"
            ) as f:
                f.writelines(
                    f"Gen:{gen_}, {self.population[fit_all_best]}, {self.fitness[fit_all_best]} \n"
                )

        return

    @staticmethod
    def get_good_bad_smiles(fitness, population, generation_size):
        """
        Given fitness values of all SMILES in population, and the generation size, 
        this function smplits  the population into two lists: keep_smiles & replace_smiles. 
        
        Parameters
        ----------
        fitness : (list of floats)
            List of floats representing properties for molecules in population.
        population : (list of SMILES)
            List of all SMILES in each generation.
        generation_size : (int)
            Number of molecules in each generation.

        Returns
        -------
        keep_smiles : (list of SMILES)
            A list of SMILES that will be untouched for the next generation. .
        replace_smiles : (list of SMILES)
            A list of SMILES that will be mutated/crossed-oved for forming the subsequent generation.

        """

        fitness = np.array(fitness)
        idx_sort = fitness.argsort()[::-1]  # Best -> Worst
        keep_ratio = 0.2
        keep_idx = int(len(list(idx_sort)) * keep_ratio)
        try:

            F_50_val = fitness[idx_sort[keep_idx]]
            F_25_val = np.array(fitness) - F_50_val
            F_25_val = np.array([x for x in F_25_val if x < 0]) + F_50_val
            F_25_sort = F_25_val.argsort()[::-1]
            F_25_val = F_25_val[F_25_sort[0]]

            prob_ = 1.0 / (3.0 ** ((F_50_val - fitness) / (F_50_val - F_25_val)) + 1)

            prob_ = prob_ / sum(prob_)
            to_keep = np.random.choice(generation_size, keep_idx, p=prob_)
            to_replace = [i for i in range(generation_size) if i not in to_keep][
                0 : generation_size - len(to_keep)
            ]

            keep_smiles = [population[i] for i in to_keep]
            replace_smiles = [population[i] for i in to_replace]

            best_smi = population[idx_sort[0]]
            if best_smi not in keep_smiles:
                keep_smiles.append(best_smi)
                if best_smi in replace_smiles:
                    replace_smiles.remove(best_smi)

            if keep_smiles == [] or replace_smiles == []:
                raise Exception("Badly sampled population!")
        except:
            keep_smiles = [population[i] for i in idx_sort[:keep_idx]]
            replace_smiles = [population[i] for i in idx_sort[keep_idx:]]

        return keep_smiles, replace_smiles

    def log(self):
        pass

    @staticmethod
    def flatten_list(nested_list):
        return [item for sublist in nested_list for item in sublist]




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Apr  4 13:50:21 2021

@author: akshat
"""
from selfies import encoder, decoder
from rdkit.Chem import MolFromSmiles as smi2mol
from rdkit.Chem import MolToSmiles as mol2smi
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit import RDLogger

RDLogger.DisableLog("rdApp.*")

import warnings
warnings.simplefilter(action="ignore", category=FutureWarning)


def sanitize_smiles(smi):
    """Return a canonical smile representation of smi
    Parameters:
    smi (string) : smile string to be canonicalized 
    Returns:
    mol (rdkit.Chem.rdchem.Mol) : RdKit mol object                          (None if invalid smile string smi)
    smi_canon (string)          : Canonicalized smile representation of smi (None if invalid smile string smi)
    conversion_successful (bool): True/False to indicate if conversion was  successful 
    """
    try:
        mol = smi2mol(smi, sanitize=True)
        smi_canon = mol2smi(mol, isomericSmiles=False, canonical=True)
        return (mol, smi_canon, True)
    except:
        return (None, None, False)


def get_frags(smi, radius):
    ''' Create fragments from smi with some radius. Remove duplicates and any
    fragments that are blank molecules.
    '''
    mol = smi2mol(smi, sanitize=True)
    frags = []
    for ai in range(mol.GetNumAtoms()):
        env = Chem.FindAtomEnvironmentOfRadiusN(mol, radius, ai)
        amap = {}
        submol = Chem.PathToSubmol(mol, env, atomMap=amap)
        frag = mol2smi(submol, isomericSmiles=False, canonical=True)
        frags.append(frag)
    return list(filter(None, list(set(frags))))

def form_fragments(smi):
    ''' Create fragments of certain radius. Returns a list of fragments
    using SELFIES characters.
    '''
    selfies_frags = []
    unique_frags = get_frags(smi, radius=3)
    for item in unique_frags:
        sf = encoder(item)
        if sf is None:
            continue
        dec_ = decoder(sf)

        try:
            m = Chem.MolFromSmiles(dec_)
            Chem.Kekulize(m)
            dearom_smiles = Chem.MolToSmiles(
                m, canonical=False, isomericSmiles=False, kekuleSmiles=True
            )
            dearom_mol = Chem.MolFromSmiles(dearom_smiles)
        except:
            continue

        if dearom_mol == None:
            raise Exception("mol dearom failes")

        selfies_frags.append(encoder(dearom_smiles))

    return selfies_frags

    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Jul 31 16:39:03 2021

@author: akshat
"""
import numpy as np
import rdkit
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.DataStructs.cDataStructs import TanimotoSimilarity
import selfies 
from selfies import encoder, decoder
import multiprocessing

from rdkit import RDLogger

from .utils import get_selfies_chars

RDLogger.DisableLog("rdApp.*")

def get_fp_scores(smiles_back, target_smi):
    """
    Given a list of SMILES (smiles_back), tanimoto similarities are calculated 
    (using Morgan fingerprints) to SMILES (target_smi). 

    Parameters
    ----------
    smiles_back : (list of str)
        List of valid SMILE strings. 
    target_smi : (str)
        Valid SMILES string. 

    Returns
    -------
    smiles_back_scores : (list of floats)
        List of figerprint similarity scores of each smiles in input list. 
    """
    smiles_back_scores = []
    target = Chem.MolFromSmiles(target_smi)
    fp_target = AllChem.GetMorganFingerprint(target, 2)
    for item in smiles_back:
        mol = Chem.MolFromSmiles(item)
        fp_mol = AllChem.GetMorganFingerprint(mol, 2)
        score = TanimotoSimilarity(fp_mol, fp_target)
        smiles_back_scores.append(score)
    return smiles_back_scores


def get_joint_sim(all_smiles, starting_smile, target_smile):
    """
    Get joint similarity values for all smiles in all_smiles, calculated with 
    refernce to starting_smile & target_smile. 

    Parameters
    ----------
    all_smiles : (list of string)
        List of SMILE strings.
    starting_smile : (str)
        Input smiles string.
    target_smile : (str)
        Input smiles string.

    Returns
    -------
    better_score : (list of floats)
        List of joint similarity scores for all smiles in all_smiles.
    """
    scores_start = get_fp_scores(all_smiles, starting_smile)  # similarity to target
    scores_target = get_fp_scores(
        all_smiles, target_smile
    )  # similarity to starting structure
    data = np.array([scores_target, scores_start])

    avg_score = np.average(data, axis=0)
    better_score = avg_score - (np.abs(data[0] - data[1]))
    better_score = (
        ((1 / 9) * better_score ** 3)
        - ((7 / 9) * better_score ** 2)
        + ((19 / 12) * better_score)
    )

    return better_score


def obtain_path(starting_smile, target_smile):
    """
    Create a single path between molecules starting_smile and target_smile. 

    Parameters
    ----------
    starting_smile : (str)
        Valid SMILES string.
    target_smile : (str)
        Valid SMILES string.

    Returns
    -------
    path_smiles : (list of str)
        List of all smiles strings encountered while creating a path.
    """
    starting_selfie = encoder(starting_smile)
    target_selfie = encoder(target_smile)

    starting_selfie_chars = get_selfies_chars(starting_selfie)
    target_selfie_char = get_selfies_chars(target_selfie)

    # Pad the smaller string
    if len(starting_selfie_chars) < len(target_selfie_char):
        for _ in range(len(target_selfie_char) - len(starting_selfie_chars)):
            starting_selfie_chars.append(" ")
    else:
        for _ in range(len(starting_selfie_chars) - len(target_selfie_char)):
            target_selfie_char.append(" ")

    indices_diff = [
        i
        for i in range(len(starting_selfie_chars))
        if starting_selfie_chars[i] != target_selfie_char[i]
    ]
    path = {}
    path[0] = starting_selfie_chars

    for iter_ in range(len(indices_diff)):
        idx = np.random.choice(indices_diff, 1)[0]  # Index to be operated on
        indices_diff.remove(idx)  # Remove that index

        # Select the last member of path:
        path_member = path[iter_].copy()

        # Mutate that character to the correct value:
        path_member[idx] = target_selfie_char[idx]
        path[iter_ + 1] = path_member.copy()

    # Collapse path to make them into SELFIE strings
    paths_selfies = []
    for i in range(len(path)):
        selfie_str = "".join(x for x in path[i])
        paths_selfies.append(selfie_str.replace(" ", ""))

    if paths_selfies[-1] != target_selfie:
        raise Exception("Unable to discover target structure!")
    path_smiles = [decoder(x) for x in paths_selfies]

    return path_smiles


def perform_crossover(comb_smi, num_random_samples):
    """
    Create multiple paths between SMILES in comb_smi to obtain median molecules, 
    representing the crossover structure. 

    Parameters
    ----------
    comb_smi : (str)
        Two smiles string concatenated using xxx (example: CCCCCCxxxSSS).
    num_random_samples : (int)
        Number of different smiles orientations to consider while forming paths. 

    Returns
    -------
    collect_smiles_canon : (list of SMILES)
        List of all potential unique median molecules enoucntered during path formation.
    """
    smi_a, smi_b = comb_smi.split("xxx")
    mol_a, mol_b = Chem.MolFromSmiles(smi_a), Chem.MolFromSmiles(smi_b)
    Chem.Kekulize(mol_a)
    Chem.Kekulize(mol_b)

    randomized_smile_orderings_a = []
    for _ in range(num_random_samples):
        randomized_smile_orderings_a.append(
            rdkit.Chem.MolToSmiles(
                mol_a,
                canonical=False,
                doRandom=True,
                isomericSmiles=False,
                kekuleSmiles=True,
            )
        )

    randomized_smile_orderings_b = []
    for _ in range(num_random_samples):
        randomized_smile_orderings_b.append(
            rdkit.Chem.MolToSmiles(
                mol_b,
                canonical=False,
                doRandom=True,
                isomericSmiles=False,
                kekuleSmiles=True,
            )
        )

    collect_smiles = []
    for smi_1 in randomized_smile_orderings_a:
        for smi_2 in randomized_smile_orderings_b:
            for item in obtain_path(smi_1, smi_2):
                collect_smiles.append(item)

    collect_smiles_canon = []
    for item in collect_smiles:
        try:
            smi_canon = Chem.MolToSmiles(
                Chem.MolFromSmiles(item, sanitize=True),
                isomericSmiles=False,
                canonical=True,
            )
            if  smi_canon != "":  # Size restriction!
                collect_smiles_canon.append(smi_canon)
        except:
            continue

    collect_smiles_canon = list(set(collect_smiles_canon))

    return collect_smiles_canon


def crossover_smiles(smiles_join, crossover_num_random_samples):
    """
    Return a list of smiles (crossover molecules) that are ordered (highest to lowest)
    by joint similarity scores. 

    Parameters
    ----------
    smiles_join : (str)
        Two smiles string concatenated using xxx (example: CCCCCCxxxSSS).

    Returns
    -------
    med_all_ord : (list of SMILES)
        List of crossover molecules that are ordered (highest to lowest)
        by joint similarity scores.
    """
    map_ = {}

    map_[smiles_join] = perform_crossover(
        smiles_join, num_random_samples=crossover_num_random_samples
    )

    # map_ordered = {}
    for key_ in map_:
        med_all = map_[key_]
        smi_1, smi_2 = key_.split("xxx")
        joint_sim = get_joint_sim(med_all, smi_1, smi_2)

        joint_sim_ord = np.argsort(joint_sim)
        joint_sim_ord = joint_sim_ord[::-1]

        med_all_ord = [med_all[i] for i in joint_sim_ord]

    return med_all_ord
    